/*
 * Copyright (c) 2004-2007 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2006 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * Copyright (c) 2010      Cisco Systems, Inc. All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */
#include "ompi_config.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif  /* HAVE_UNISTD_H */
#ifdef HAVE_STRING_H
#include <string.h>
#endif  /* HAVE_STRING_H */
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif  /* HAVE_SYS_TYPES_H */
#include <sys/mman.h>
//TODO: this will be in linux devkit

#if OMPI_BTL_HAVE_SRIO
#include <linux/rio_mport_cdev.h>
#include <linux/rio_regs.h>
#endif

#include "ompi/constants.h"
#include "opal/runtime/opal.h"
#include "opal/mca/event/event.h"
#include "opal/mca/base/mca_base_param.h"
#include "btl_srio.h"
#include "btl_srio_frag.h"
#include "btl_srio_fifo.h"

#include "pktfwdK2H/mpi_interfaces.h"

static int mca_btl_srio_component_progress (void);
static int srio_register(void);

/*
 * Shared Memory (SELF) component instance. 
 */

mca_btl_srio_component_t mca_btl_srio_component = {
    {  /* super is being filled in */
        /* First, the mca_base_component_t struct containing meta information
          about the component itself */
        {
            MCA_BTL_BASE_VERSION_2_0_0,

            "srio", /* MCA component name */
            OMPI_MAJOR_VERSION,  /* MCA component major version */
            OMPI_MINOR_VERSION,  /* MCA component minor version */
            OMPI_RELEASE_VERSION,  /* MCA component release version */
            mca_btl_srio_component_open,  /* component open */
            mca_btl_srio_component_close,  /* component close */
            NULL,
            srio_register,
        },
        {
            /* The component is checkpoint ready */
            MCA_BASE_METADATA_PARAM_CHECKPOINT
        },

        mca_btl_srio_component_init, 
        mca_btl_srio_component_progress,
    }  /* end super */
};
/**************************************************************************************/
static char* mca_btl_srio_param_register_string(
    const char* param_name,
    const char* default_value)
{
    char *param_value;
    int id = mca_base_param_register_string("btl","srio",param_name,NULL,default_value);
    mca_base_param_lookup_string(id, &param_value);
    return param_value;
}

static int mca_btl_srio_param_register_int(
    const char* param_name,
    int default_value)
{
    int id = mca_base_param_register_int("btl","srio",param_name,NULL,default_value);
    int param_value = default_value;
    mca_base_param_lookup_int(id,&param_value);
    return param_value;
}
/**************************************************************************************/
static int srio_register(void)
{
    int i;
    FILE *node_id_detect, *cart_id_detect;
    unsigned char node_id_buff[8], cart_id_buff[8];
#if OMPI_BTL_HAVE_SRIO
    /*********************************************************************/
    /** Find location in Slayton topology: if this is EVM, default to 1 **/
    /*********************************************************************/
    node_id_detect = fopen("/proc/device-tree/sl1500/node/id", "rb");
    if(node_id_detect)
    {
       fread(node_id_buff, 1, 2, node_id_detect);
       mca_btl_srio_component.node_id = strtoul(node_id_buff, (char **)NULL, 8);
       fclose (node_id_detect);
    } else mca_btl_srio_component.node_id = 0;
    /** Find location in Slayton topology: if this is EVM, default to 1 **/
    cart_id_detect = fopen("/proc/device-tree/sl1500/cartridge/id", "rb");
    if(cart_id_detect)
    {
       fread(cart_id_buff, 1, 2, cart_id_detect);
       mca_btl_srio_component.cartridge_id = strtoul(cart_id_buff, (char **)NULL, 8);
       fclose (cart_id_detect);
    } else mca_btl_srio_component.cartridge_id = 0;

    mca_btl_srio_component.srio_id = 1;
  
    mca_btl_srio_component.ddr_loc_addr =
        mca_btl_srio_param_register_int("ddr_loc_addr", BTL_SRIO_DDR_ADDRESS);
    mca_btl_srio_component.msmc_loc_addr =
        mca_btl_srio_param_register_int("msmc_loc_addr", BTL_SRIO_MSMC_ADDRESS);

    mca_btl_srio_component.fifo_size =
        mca_btl_srio_param_register_int("fifo_size", BTL_SRIO_MAX_FIFO_ITEMS);

    /* register SELF component parameters */
    mca_base_param_reg_int( (mca_base_component_t*)&mca_btl_srio_component, "free_list_num",
                            "Number of fragments by default", false, false,
                            0, &mca_btl_srio_component.free_list_num );
    mca_base_param_reg_int( (mca_base_component_t*)&mca_btl_srio_component, "free_list_max",
                            "Maximum number of fragments", false, false,
                            -1, &mca_btl_srio_component.free_list_max );
    mca_base_param_reg_int( (mca_base_component_t*)&mca_btl_srio_component, "free_list_inc",
                            "Increment by this number of fragments", false, false,
                            32, &mca_btl_srio_component.free_list_inc );

    mca_btl_srio_component.hostnameList = SrioHostnameListCreate (BTL_SRIO_PHYS_HOSTS);
    if(mca_btl_srio_component.hostnameList == NULL) {
       opal_output_verbose(10, mca_btl_base_output, "BTL_SRIO: host to physical location file %s cannot be opened\n", BTL_SRIO_PHYS_HOSTS);
       return OMPI_SUCCESS;
    }

    if(BTL_SRIO_EVM(mca_btl_srio_component)) {
      char my_hostname[80];
      gethostname(my_hostname, 80);
      if(SrioHostname2PhysID (mca_btl_srio_component.hostnameList, my_hostname, 
                              &mca_btl_srio_component.cartridge_id, &mca_btl_srio_component.node_id) < 0)
      {
        opal_output(0, "BTL_SRIO: Cannot find physical location of hostname %s, in list %s", my_hostname, BTL_SRIO_PHYS_HOSTS);
        return OMPI_ERR_OUT_OF_RESOURCE;
      }  
    } 
    strncpy (mca_btl_srio_component.routing_tbl_filename,
             mca_btl_srio_param_register_string("srio_routing_tbl", BTL_SRIO_ROUTING_TBL_FILE), 320);
    opal_output_verbose(20, mca_btl_base_output, "BTL_SRIO: Routing table filename:%s cartridge_id=%d node_id=%d\n", 
      mca_btl_srio_component.routing_tbl_filename, mca_btl_srio_component.cartridge_id, mca_btl_srio_component.node_id);
    /** Import routing table **/
    mca_btl_srio_component.routingTbl = SrioRoutingStructuresCreate(mca_btl_srio_component.routing_tbl_filename); 
    if(mca_btl_srio_component.routingTbl == NULL) {
       opal_output_verbose(10, mca_btl_base_output, "BTL_SRIO: Routing table %s cannot be opened\n", mca_btl_srio_component.routing_tbl_filename);
       return OMPI_ERR_OUT_OF_RESOURCE;
    }
#endif
    return OMPI_SUCCESS;
}
/*
 *  Called by MCA framework to open the component, registers
 *  component parameters.
 */

int mca_btl_srio_component_open(void)
{
#if OMPI_BTL_HAVE_SRIO
    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: srio_component_open!!!\n");
    mca_btl_srio.super.btl_exclusivity = MCA_BTL_EXCLUSIVITY_HIGH;
    mca_btl_srio.super.btl_eager_limit = BTL_SRIO_EAGER_SIZE;
    mca_btl_srio.super.btl_rndv_eager_limit = BTL_SRIO_EAGER_SIZE;
    mca_btl_srio.super.btl_max_send_size = BTL_SRIO_MAX_SEND_SIZE;
    mca_btl_srio.super.btl_flags = MCA_BTL_FLAGS_SEND | MCA_BTL_FLAGS_SEND_INPLACE;
    mca_btl_srio.super.btl_seg_size = sizeof (mca_btl_base_segment_t);
    mca_btl_srio.super.btl_bandwidth = 10000;
    mca_btl_srio.super.btl_latency = 1;
    mca_btl_base_param_register(&mca_btl_srio_component.super.btl_version, &mca_btl_srio);
    /**************************************/
    mca_btl_srio_component.max_frag_size = mca_btl_srio.super.btl_max_send_size;
    mca_btl_srio_component.eager_limit = mca_btl_srio.super.btl_eager_limit;
    /**************************************/
    /* initialize objects */
    OBJ_CONSTRUCT(&mca_btl_srio_component.srio_lock, opal_mutex_t);
    OBJ_CONSTRUCT(&mca_btl_srio_component.srio_frags_eager, ompi_free_list_t);
    OBJ_CONSTRUCT(&mca_btl_srio_component.srio_frags_max, ompi_free_list_t);
#endif
    return OMPI_SUCCESS;
}
/********************************************************************************************************************/
/*
 * component cleanup - sanity checking of queue lengths
 */

int mca_btl_srio_component_close(void)
{
#if OMPI_BTL_HAVE_SRIO
    opal_output_verbose(20, mca_btl_base_output, "\nTIMPIDBG: srio_component_close!!!\n");
    OBJ_DESTRUCT(&mca_btl_srio_component.srio_lock);
    OBJ_DESTRUCT(&mca_btl_srio_component.srio_frags_eager);
    OBJ_DESTRUCT(&mca_btl_srio_component.srio_frags_max);
#endif
    return OMPI_SUCCESS;
}
/***************************************************************************
      Generate file descriptors per port, so that they could be used for file transfer.
      The /dev/rio_mport1-4 may be numbered randomly, depending upon the order of bringup
      of neighbouring nodes. To find the port associated with a /dev/rio_mport<n> an ioctl() API
       needs to be invoked, which returns the SRIO port corresponding to that /dev/rio_mport<n> file.
       This routine below, collects that and saves the file descriptos in an array, indexed by port number, so
       that it could be used at run-time.
*****************************************************************************/
int mca_btl_srio_mport_init(int fd_per_port[])
{
   char mport_files[BTL_SRIO_MAX_PORTS][20]={ "/dev/rio_mport0",
                                              "/dev/rio_mport1",
                                              "/dev/rio_mport2",
                                              "/dev/rio_mport3" };
   int mport_fd, port, res, tot_ports = 0;
   unsigned int port_idx;
#if OMPI_BTL_HAVE_SRIO
   for(port = 0; port < BTL_SRIO_MAX_PORTS; port++) fd_per_port[port] = -1; 
   
   for(port = 0, tot_ports = 0; port < BTL_SRIO_MAX_PORTS; port++) 
   {
     opal_output_verbose(100, mca_btl_base_output, "BTL SRIO mport_init attempt to open %s", mport_files[port]);
     mport_fd = open(mport_files[port], 0, 0);
     if (mport_fd < 0) continue;

     port_idx=0xFFFFFFFF;
     res = ioctl(mport_fd, RIO_MPORT_MAINT_PORT_IDX_GET, &port_idx);
     opal_output_verbose(100, mca_btl_base_output, "Port index of %s is %d\n", mport_files[port], port_idx);
     if(port_idx!=0xFFFFFFFF) {
        fd_per_port[port_idx]=mport_fd;
     }
     tot_ports ++;
     opal_output_verbose(10, mca_btl_base_output, "mport_init %d->%d, fd=%d\n", port, port_idx, mport_fd);
   }
   opal_output_verbose(10, mca_btl_base_output, "mport_total_ports=%d\n", tot_ports);
#endif
   return tot_ports;
}
/********************************************************************************************************************/
/*
 *  SRIO component initialization
 */
mca_btl_base_module_t** mca_btl_srio_component_init( int *num_btls, 
                                                     bool enable_progress_threads,
                                                     bool enable_mpi_threads )
{
    mca_btl_base_module_t **btls = NULL;
    char *ptr_hostname;
    int rc, i, tot_ports;
#if OMPI_BTL_HAVE_SRIO
    /** mpm_transport init **/
    mpm_transport_open_t ocfg = {
      .open_mode      = (O_SYNC|O_RDWR),
      .msec_timeout   = 2000
    };

    mpm_transport_mmap_t mcfg = {
      .mmap_prot      = (PROT_READ|PROT_WRITE),
      .mmap_flags     = MAP_SHARED,
    };
    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: srio_component_init!!!");
    /* allocate the Shared Memory PTL */
    *num_btls = 1;
    btls = (mca_btl_base_module_t**)malloc((*num_btls)*sizeof(mca_btl_base_module_t*));
    if (NULL == btls) {
        return NULL;
    }
    if(!BTL_SRIO_EVM(mca_btl_srio_component)) 
    {
      if(mca_btl_srio_component.routingTbl == NULL) {
        opal_output(0, "BTL_SRIO: Routing table is not valid\n");
        return NULL;
      }
    } else {
      opal_output_verbose(10, mca_btl_base_output, "BTL_SRIO: Host to physical location file is not valid\n");
      return NULL;
    }
    mca_btl_srio_component.num_srio_btls = 1;
    mca_btl_srio_component.srio_btls     = btls;

    /* initialize free lists */
    rc = ompi_free_list_init_ex_new (&mca_btl_srio_component.srio_frags_eager,
                                     sizeof(mca_btl_srio_frag1_t),
                                     opal_cache_line_size, OBJ_CLASS(mca_btl_srio_frag1_t),
                                     mca_btl_srio.super.btl_eager_limit,
                                     opal_cache_line_size,
                                     mca_btl_srio_component.free_list_num,
                                     mca_btl_srio_component.free_list_max,
                                     mca_btl_srio_component.free_list_inc,
                                     NULL, (ompi_free_list_item_init_fn_t) mca_btl_srio_frag_init,
                                     (void *) &mca_btl_srio);

    if (OPAL_UNLIKELY(OMPI_SUCCESS != rc)) {
        opal_output(0, "error creating srio fragment1 free list");
        return rc;
    }

    rc = ompi_free_list_init_ex_new (&mca_btl_srio_component.srio_frags_max,
                                     sizeof(mca_btl_srio_frag1_t),
                                     opal_cache_line_size, OBJ_CLASS(mca_btl_srio_frag1_t),
                                     mca_btl_srio.super.btl_max_send_size,
                                     opal_cache_line_size,
                                     mca_btl_srio_component.free_list_num,
                                     mca_btl_srio_component.free_list_max,
                                     mca_btl_srio_component.free_list_inc,
                                     NULL, (ompi_free_list_item_init_fn_t) mca_btl_srio_frag_init,
                                     (void *) &mca_btl_srio);

   if (OPAL_UNLIKELY(OMPI_SUCCESS != rc)) {
      opal_output(0, "error creating srio fragment2 free list");
      return rc;
   }

   /* get pointer to the btls */
   btls[0] = (mca_btl_base_module_t *)(&mca_btl_srio);

   /*******************************************/
   /** Open all available MPORT interfaces   **/
   /*******************************************/
   if(mca_btl_srio_mport_init (mca_btl_srio_component.mport_fd) <= 0) return NULL; /* in case no port available */
   /************************************************************************************************************************/
   /** SHARED MEMORY ACCESS: one MSMC region is mapped for both interfaces, and two DDR regions, one per HYPLNK interface **/
   /************************************************************************************************************************/

   /** Open local shared memory interface **/
   mca_btl_srio_component.mpm_transport_shmem = (void *)mpm_transport_open("arm-local-shmem", &ocfg);
   if (!mca_btl_srio_component.mpm_transport_shmem) {
      opal_output(0, "BTL SRIO shmem open failed");
      return NULL;
   }
   opal_output_verbose(20, mca_btl_base_output, "BTL SRIO shmem open successfull!!");
   /****---------------------------------****/
   /** Map local msmc region to user space **/
   /****---------------------------------****/
   mca_btl_srio_component.win_msmc = mpm_transport_mmap((mpm_transport_h)mca_btl_srio_component.mpm_transport_shmem, 
                                                         mca_btl_srio_component.msmc_loc_addr, BTL_SRIO_MSMC_MMAP_SIZE, &mcfg);
   if(!mca_btl_srio_component.win_msmc)
   {
     opal_output(0, "BTL SRIO mmap_failed_win_msmc!");
     return NULL;
   }
   memset (mca_btl_srio_component.win_msmc, 0, BTL_SRIO_MSMC_MMAP_SIZE);

   opal_output_verbose(20, mca_btl_base_output, "BTL SRIO shmem MSMC mmap successfull!!");
   /** Map local DDR region to user space **/
   mca_btl_srio_component.win_ddr = mpm_transport_mmap((mpm_transport_h)mca_btl_srio_component.mpm_transport_shmem,
                                                         mca_btl_srio_component.ddr_loc_addr, BTL_SRIO_MMAP_SIZE, &mcfg);
   if(!mca_btl_srio_component.win_ddr)
   {
      opal_output (0, "BTL SRIO mmap_failed_win_ddr!");
      return NULL;
   }
   opal_output_verbose(20, mca_btl_base_output, "BTL SRIO shmem DDR mmap successfull!!");
   opal_output_verbose(10, mca_btl_base_output, "BTL SRIO turned ON !!!");
   return btls;
#else
   opal_output(0, "SRIO support not built in!!!\n");
   return NULL;
#endif
}
/*******************************************************************************************/
void mca_btl_srio_one_fragment_received (mca_btl_srio_hdr_t *hdr)
{
    mca_btl_srio_frag_t Frag;
    mca_btl_active_message_callback_t* reg;

    /* recv upcall */
    reg = mca_btl_base_active_message_trigger + hdr->tag;
    Frag.base.des_dst = &Frag.segment;
    Frag.segment.seg_addr.pval = ((char *)hdr) + sizeof(mca_btl_srio_hdr_t);

    memcpy (&Frag.segment.seg_len, &hdr->len, sizeof(size_t));
    Frag.base.des_dst_cnt = 1;
    reg->cbfunc(&mca_btl_srio.super, hdr->tag, &(Frag.base), reg->cbdata);
    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG:SRIO component_progress FRAG RCV end");
}
/*******************************************************************************************/
static int mca_btl_srio_component_progress (void)
{
    mca_btl_srio_hdr_t *hdr;
    int peer_rank, rc = 0, nevents = 0;
    int j, ret_val;
    int ack_frag = 0;
    char *ptr;

#if OMPI_BTL_HAVE_SRIO
    for(j = 0; j < mca_btl_srio_component.endpoint_count; j ++)
    { // Scan all SRIO ports
      if(j == mca_btl_srio_component.my_rank) continue;
      ret_val = srio_fifo_read_begin (mca_btl_srio_component.rd_fifo[j]);
        if(ret_val == -1) continue;
        opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG component_progress reading FIFOs A%d ... fragment at %d", j, ret_val);
        hdr = (mca_btl_srio_hdr_t *)(mca_btl_srio_component.win_ddr + j * BTL_SRIO_MAX_SEND_SIZE * BTL_SRIO_MAX_FIFO_ITEMS + ret_val * BTL_SRIO_MAX_SEND_SIZE);
        mca_btl_srio_one_fragment_received (hdr);
        opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG component_progress reading FIFOs B%d", j);
      srio_fifo_read_end(mca_btl_srio_component.rd_fifo[j]);
      opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG component_progress reading FIFOs C%d", j);
      nevents ++;
      opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG component_progress reading FIFOs mapper over SRIO END(%d)", j);
    }
#endif
    return 1;
}
/*** nothing past this point ***/
