#include "ompi_config.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stddef.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <getopt.h>
#include <time.h>


#if OMPI_BTL_HAVE_SRIO
#include <linux/rio_mport_cdev.h>
#include <linux/rio_regs.h>
#endif


#include "opal/class/opal_bitmap.h"
#include "opal/datatype/opal_convertor.h"
#include "opal/sys/atomic.h"

#include "btl_srio.h"
#include "btl_srio_fifo.h"

#define RIO_DBELL_MASK 0x1f

/* Direct I/O modes */
#define RIO_DIO_MODE_READ    0x0000
#define RIO_DIO_MODE_WRITER  0x0001
#define RIO_DIO_MODE_WRITE   0x0002
#define RIO_DIO_MODE_SWRITE  0x0003

/***************************************************************************************************/
static int64_t clock_diff (struct timespec *start, struct timespec *end)
{
    return (end->tv_sec - start->tv_sec) * 1000000000
             + end->tv_nsec - start->tv_nsec;
}
/*****************************************************************************************************************/
void *srio_read_fifo_init(void *fifo_address, unsigned int base_payload_addr, unsigned int fragment_size, int fifo_size)
{
    int qsize;
    srio_fifo_t *fifo;
    
    memset (fifo_address, 0, sizeof(srio_fifo_t));
    fifo = (srio_fifo_t *)fifo_address;
    fifo->base_payload_addr = base_payload_addr;
    fifo->fragment_size     = fragment_size;
    fifo->qsize             = fifo_size;
    msync(fifo, sizeof(srio_fifo_t), MS_SYNC); 
    return fifo_address;
}
/*****************************************************************************************************************/

void *srio_write_fifo_init (int fd, int destid, unsigned int remote_addr)
{
  srio_write_fifo_t *loc_fifo = malloc(sizeof(srio_write_fifo_t));
  memset (loc_fifo, 0, sizeof(srio_write_fifo_t));
  loc_fifo->remote_addr = remote_addr;
  loc_fifo->destid = destid;
  loc_fifo->fd = fd;
  return loc_fifo;
}

/*****************************************************************************************************************/
int srio_fifo_write(void *loc_fifo_p, void *buffer, int len)
{
#if OMPI_BTL_HAVE_SRIO
    int fifo_qsize;
    int head_cnt;
    int tail_cnt;
    int fifo_entries;
    struct rio_mport_dio_transfer dio;
    int cmd, res;
#ifdef BTL_SRIO_VERBOSE
//BENCHMARK
    struct timespec tp_start, tp_end;
#endif
    srio_write_fifo_t *loc_fifo = (srio_write_fifo_t *)loc_fifo_p;
    dio.id = loc_fifo->destid;
    opal_output_verbose(20, mca_btl_base_output, "SRIO_FIFO_WRITE to srio_id %d (rmt %08x) head %d tail %d\n", dio.id, 
         loc_fifo->remote_addr, loc_fifo->rmt_fifo.head_cnt, loc_fifo->rmt_fifo.tail_cnt);
    if(!loc_fifo->update_cnt || ((loc_fifo->rmt_fifo.head_cnt - loc_fifo->rmt_fifo.tail_cnt) >= loc_fifo->rmt_fifo.qsize))
    {
      dio.tgt_addr = (uint64_t)loc_fifo->remote_addr;
      dio.src_addr = &(loc_fifo->rmt_fifo);
      dio.length = sizeof(srio_fifo_t);
      dio.mode = RIO_DIO_MODE_READ;
      cmd = RIO_MPORT_DIO_TRANSFER;
      res = ioctl(loc_fifo->fd, cmd, &dio);
      if (res < 0) {
        opal_output(0, "ioctl READ DIO_TRANSFER from id%d failed [fd=%d] fromAddr:%08llx toAddr:(%p) len:%08x ", 
          dio.id, loc_fifo->fd, dio.tgt_addr, dio.src_addr, dio.length);
        close(loc_fifo->fd);
        return(-1);
      }
      loc_fifo->update_cnt ++;
      opal_output_verbose(20, mca_btl_base_output, "SRIO_FIFO_WRITE to srio_id %d UPDATE_cnt %d head %d tail %d qsize %d\n", 
         dio.id, loc_fifo->update_cnt, loc_fifo->rmt_fifo.head_cnt, loc_fifo->rmt_fifo.tail_cnt, loc_fifo->rmt_fifo.qsize);
    }
    /* if there is no free slot to write, report exhausted resource */

    if((loc_fifo->rmt_fifo.head_cnt - loc_fifo->rmt_fifo.tail_cnt)  >= loc_fifo->rmt_fifo.qsize) {
      return -1;
    }
    dio.tgt_addr = loc_fifo->rmt_fifo.base_payload_addr + (loc_fifo->rmt_fifo.head_cnt % loc_fifo->rmt_fifo.qsize) * loc_fifo->rmt_fifo.fragment_size;
    dio.src_addr = buffer;
    dio.length = len;
    dio.mode = RIO_DIO_MODE_WRITER;
    cmd = RIO_MPORT_DIO_TRANSFER;
#ifdef BTL_SRIO_VERBOSE
//BENCHMARK
clock_gettime(CLOCK_MONOTONIC, &tp_start);
//BENCHMARK
#endif
    res = ioctl(loc_fifo->fd, cmd, &dio);
    if (res < 0) {
       opal_output(0, "ioctl WRITE DIO_TRANSFER 1 failed [fd=%d]", loc_fifo->fd);
       close(loc_fifo->fd);
       return(-1);
    }
#ifdef BTL_SRIO_VERBOSE
//BENCHMARK
clock_gettime(CLOCK_MONOTONIC, &tp_end);
if(len >= 8192) {
  opal_output(0, "BTL SRIO ioctl write: %lld ns for %d bytes", clock_diff(&tp_start, &tp_end), len);
  clock_gettime(CLOCK_MONOTONIC, &tp_start);
  clock_gettime(CLOCK_MONOTONIC, &tp_end);
  opal_output(0, "BTL SRIO ioctl measurement: %lld ns", clock_diff(&tp_start, &tp_end));
}
//BENCHMARK
#endif
    /* write back, to remote memory, updated head_cnt */
    loc_fifo->rmt_fifo.head_cnt ++;
    dio.tgt_addr = loc_fifo->remote_addr + offsetof(srio_fifo_t, head_cnt);
    dio.src_addr = &loc_fifo->rmt_fifo.head_cnt;
    dio.length = sizeof(unsigned int);
    dio.mode = RIO_DIO_MODE_WRITER;
    cmd = RIO_MPORT_DIO_TRANSFER;
    res = ioctl(loc_fifo->fd, cmd, &dio);
    if (res < 0) {
       opal_output(0, "ioctl WRITE DIO_TRANSFER 2 failed [fd=%d]",  loc_fifo->fd);
       close(loc_fifo->fd);
       return(-1);
    }
#endif
    return 0;
}
/*****************************************************************************************************************/

int srio_fifo_read_begin(void *fifo_ptr)
{
    int fifo_qsize;
    int head_cnt, tail_cnt;
    srio_fifo_t *fifo = (srio_fifo_t *)fifo_ptr;
    int value = -1;

    msync(fifo, sizeof(srio_fifo_t), MS_INVALIDATE); 
    head_cnt = fifo->head_cnt;
    tail_cnt = fifo->tail_cnt;
    /* if there is no free slot to write, report exhausted resource */
    if((head_cnt - tail_cnt) > 0) {
      value = (tail_cnt % fifo->qsize);
    } 
    return value;
}
/*****************************************************************************************************************/

void srio_fifo_read_end(void *fifo_ptr)
{
    srio_fifo_t *fifo = (srio_fifo_t *)fifo_ptr;
    int tail_cnt;

    msync(fifo, sizeof(srio_fifo_t), MS_INVALIDATE); 
    tail_cnt = fifo->tail_cnt;
    tail_cnt ++;
    fifo->tail_cnt = tail_cnt;
    msync(fifo, sizeof(srio_fifo_t), MS_SYNC); 
}
/*** nothing past this point ***/
